from typing import Dict, List
from yolink_model import BRDP
from yolink_auth_mgr import YoLinkAuthMgr
from yolink_device import YoLinkDeviceEntry
from .const import YOLINK_API_MQTT_BROKER, YOLINK_API_MQTT_BROKER_POER
import uuid
import logging
import asyncio


__LOGGER = logging.getLogger(__name__)


class HomeEventSub:
    """Home Event subscription."""

    def __init__(self, homeId: str):
        """Init Home Subscription."""
        self.homeId = homeId
        self.platformDeviceSubscription: Dict[str,
                                              List[YoLinkDeviceEntry]] = {}

    def attachPlatformDevices(self, platform: str, devices: List[YoLinkDeviceEntry]):
        """Relate YoLink devices to this subscription."""
        self.platformDeviceSubscription[platform] = devices
        return None

    def onDeviceMessage(self, device_id: str, msg: BRDP):
        """Call when message from device received."""
        for platform, devices in self.platformDeviceSubscription.items():
            for device in devices:
                if device.device_id == device_id:
                    device.push_data(msg)
        return None


class MqttSub:
    """Mqtt subscription."""

    def __init__(self, *topics) -> None:
        """Init Mqtt subscription"""
        self.sub_topics = topics
        self.sub_topics_success = {}

    def on_topic_sub_success(self, topic: str):
        """On MQTT Sub Success."""
        self.sub_topics_success[topic] = True

    def on_mqtt_disconnect(self):
        """On Mqtt disconnected."""
        self.sub_topics_success = {}

    def subscribe(self):
        """Return topics need to subscribe."""
        ret = []
        for topic in self.sub_topics:
            if not (
                topic in self.sub_topics_success
                and self.sub_topics_success[topic] is True
            ):
                ret.append(topic)
        return ret

    def on_message(self, topic: str, msg: str):
        """Implement callback for MQTTClient.on_message."""
        pass


class HomeEventMqttSub(HomeEventSub, MqttSub):
    """Home Event Mqtt Subscription."""

    def __init__(self, homeId: str):
        """Init MQTT Home Subscription."""

        HomeEventSub.__init__(self, homeId)
        self.device_report_topic = f"yl-home/{self.homeId}/+/report"
        MqttSub.__init__(self, self.device_report_topic)

    def on_message(self, topic: str, msg: str):
        """On Receive MQTT Message."""
        keys = topic.split("/")
        if len(keys) == 4 and keys[3] == "report":
            self.onDeviceMessage(keys[2], BRDP.parse_raw(msg))
        return None


class MqttClient:
    """Mqtt Client"""

    def __init__(self, authManager: YoLinkAuthMgr) -> None:
        self._subscriptions: List[MqttSub] = []
        self._auth_mgr = authManager
        self.loop = asyncio.get_running_loop()
        self.init_client()

    def subHome(self, subscription: MqttSub):
        """Subscription Home."""
        self._subscriptions.append(subscription)
        return True

    def init_client(self):
        """Initialize paho mqtt client."""
        from paho.mqtt import client as mqtt
        client_id = mqtt.base62(uuid.uuid4().int, padding=22)
        self._mqttc = mqtt.Client(client_id, mqtt.MQTTv31)

        # enable mqtt log
        self._mqttc.enable_logger()
        self._mqttc.on_connect = self._mqtt_on_connection
        self._mqttc.on_disconnect = self._mqtt_on_disconnection
        self._mqttc.on_message = self._mqtt_on_message

    async def async_connect(self) -> None:
        """Connect to Mqtt broker."""
        # check and fresh access token
        await self._auth_mgr.check_and_refresh_token()
        # set mqtt
        import paho.mqtt.client as mqtt
        self._mqttc.username_pw_set(self._auth_mgr.access_token(), "")
        result = None
        try:
            result = await self.loop.run_in_executor(
                None, self._mqttc.connect, YOLINK_API_MQTT_BROKER, YOLINK_API_MQTT_BROKER_POER, 600)
        except OSError as err:
            __LOGGER.error(
                "Failed to connect to MQTT server due to exception: %s", err)
        if result is not None and result != 0:
            __LOGGER.error(
                "Failed to connect to MQTT server: %s", mqtt.error_string(
                    result)
            )
        self._mqttc.loop_start()

    def async_disconnect(self):
        """Disconnect from mqtt server."""
        def stop():
            """Stop the MQTT client."""
            # Do not disconnect, we want the broker to always publish will
            self._mqttc.loop_stop()
        self.loop.run_in_executor(None, stop)

    def _mqtt_on_connection(self, _mqttc, _userdata, _flags, result_code: int):
        """On connect callback.
        Resubscribe to all topics we were subscribed to and publish birth
        message.
        """
        import paho.mqtt.client as mqtt
        if result_code != mqtt.CONNACK_ACCEPTED:
            __LOGGER.error(
                "Unable to connect to the MQTT broker: %s",
                mqtt.connack_string(result_code),
            )
            return
        self.connected = True
        __LOGGER.info(
            "Connected to MQTT server %s:%s (%s)",
            YOLINK_API_MQTT_BROKER,
            YOLINK_API_MQTT_BROKER_POER,
            result_code,
        )
        # Group subscriptions to only re-subscribe once for each topic.
        for sub in self._subscriptions:
            for topic in sub.subscribe():
                self._mqttc.subscribe(topic, 0)

    def _mqtt_on_disconnection(self, _mqttc, _userdata, result_code):
        """Mqtt On disconnection."""
        try:
            access_token = asyncio.run_coroutine_threadsafe(
                self._auth_mgr.check_and_refresh_token(), self.loop
            ).result()
            _mqttc.username_pw_set(access_token, "")
        except Exception as e:
            print(e)

    def _mqtt_on_message(self, msg) -> None:
        """Mqtt on message."""
        __LOGGER.debug(
            "Received message on %s%s: %s",
            msg.topic,
            " (retained)" if msg.retain else "",
            msg.payload[0:8192],
        )
        for sub in self._subscriptions:
            sub.on_message(msg.topic, msg.payload.decode("UTF-8"))
